#include "atomic_block.hpp"

Atomic::Atomic(std::vector<uint32_t> inputs, std::vector<uint32_t> outputs, std::vector<uint32_t> nodeIndexList, std::string truthTable, std::vector<std::vector<uint32_t>> fanoutsVec)
{
    this->inputs = inputs;
    sort(this->inputs.begin(), this->inputs.end());
    this->outputs = outputs;
    sort(this->outputs.begin(), this->outputs.end());
    this->nodeIndexList = nodeIndexList;
    sort(this->nodeIndexList.begin(), this->nodeIndexList.end());
    this->truthTable = truthTable;
    this->fanoutsVec = fanoutsVec;
}

//***************************************************************************
Atomic::Atomic(std::vector<Cutt> cuts)
{
    //Cuts has the same inputs then we use the input vector of one of them
    this->inputs = cuts[0].returnInputs();
    for (auto i : cuts)
    {
        this->truthTable += i.truthTable;
        this->outputs.push_back(i.output);
        this->fanoutsVec.push_back(i.fanouts);
        this->nodeIndexList.insert(this->nodeIndexList.end(), i.nodeIndexList.begin(), i.nodeIndexList.end());
    }
    std::sort(this->nodeIndexList.begin(), this->nodeIndexList.end());
    this->nodeIndexList.erase(std::unique(this->nodeIndexList.begin(), this->nodeIndexList.end()), this->nodeIndexList.end());
}

//***************************************************************************
std::vector<uint32_t> Atomic::returnInputs() const
{
    return this->inputs;
}

//***************************************************************************
std::vector<uint32_t> Atomic::returnOutputs() const
{
    return this->outputs;
}

//***************************************************************************
std::string Atomic::returnTruthTable() const
{
    return this->truthTable;
}

//***************************************************************************
std::vector<uint32_t> Atomic::returnNodeIndexes() const
{
    return this->nodeIndexList;
}

//***************************************************************************
std::ostream &operator<<(std::ostream &output, const Atomic &A)
{
    for (auto i : A.outputs)
    {
        output << i << " ";
    }
    output << "---> ";
    for (auto i : A.inputs)
    {
        output << i << " ";
    }
    output << "[ ";
    for (auto i : A.nodeIndexList)
    {
        output << i << " ";
    }
    output << "]";
    output << " ===> " << A.truthTable;
    return output;
}

//***************************************************************************
std::pair<std::vector<Atomic>, std::vector<Atomic>> Atomic::identifyHAsFAs(std::vector<Cutt *> cuts)
{
    //adding the Sum cuts to an unordered_map and the Carry cuts to a vector
    //keys are the input signatures
    //we use another unordered_map to keep track of the existing signatures

    std::vector<Cutt *> HA_C_cuts, FA_C_cuts;
    std::unordered_map<std::string, Cutt *> HA_S_cuts_map, FA_S_cuts_map;
    std::unordered_map<std::string, bool> HA_S_cut_trace, FA_S_cut_trace;

    std::string signature;

    for (auto i : cuts)
    {
        signature = i->inputSig;
        switch (i->identifyCut())
        {
        case HA_C:
            HA_C_cuts.push_back(i);
            break;
        case FA_C:
            FA_C_cuts.push_back(i);
            break;
        case HA_S:
            HA_S_cuts_map[signature] = i;
            HA_S_cut_trace[signature] = true;
            break;
        case FA_S:
            FA_S_cuts_map[signature] = i;
            FA_S_cut_trace[signature] = true;
            break;
        }
    }
    std::vector<Atomic> HAs;
    std::vector<Atomic> FAs;

    for (auto i : HA_C_cuts)
    {
        signature = i->inputSig;

        if (HA_S_cut_trace[signature] == true)
        {
            Atomic HA({*HA_S_cuts_map[signature], *i});
            //we should check if HA has exactly two outputs
            //if fanoutsVec[0] or fanoutsVec[1] size is equal to zero, it means that the output of HA is also AIG's primary output. we should also consider this case!
            if ((!std::includes(HA.nodeIndexList.begin(), HA.nodeIndexList.end(), HA.fanoutsVec[0].begin(), HA.fanoutsVec[0].end()) || HA.fanoutsVec[0].empty()) && (!std::includes(HA.nodeIndexList.begin(), HA.nodeIndexList.end(), HA.fanoutsVec[1].begin(), HA.fanoutsVec[1].end()) || HA.fanoutsVec[1].empty()))
                HAs.push_back(HA);        }
    }

    for (auto i : FA_C_cuts)
    {
        signature = i->returnInputSig();
        if (FA_S_cut_trace[signature] == true)
        {
            Atomic FA({*FA_S_cuts_map[signature], *i});
            FAs.push_back(FA);
        }
    }

    for (auto i : cuts) //freeing memory
        delete i;

    return {HAs, FAs};
}

//***************************************************************************
void Atomic::removeHAsInFAs(std::pair<std::vector<Atomic>, std::vector<Atomic>> &atomicsPair)
{
    std::unordered_map<uint32_t, int> FAs_nodes_trace;
    std::unordered_map<uint32_t, bool> FAs_outputs_trace;
    std::unordered_map<uint32_t, int> FAs_nodes_indexes;

    std::unordered_map<int, bool> FA_remove_indexes; //to keep the indexex that should not be included in the FA vector

    for (int index = 0; index < atomicsPair.second.size(); index++)
    {
        for (auto i : atomicsPair.second[index].nodeIndexList)
        {
            FAs_nodes_trace[i] = index+1; //if FAs_nodes_trace[i]==0 then the node is not part of a FA
            FAs_nodes_indexes[i] = index;
        }
        for (auto i : atomicsPair.second[index].outputs)
            FAs_outputs_trace[i] = true;
    }

    std::vector<Atomic> pronedHAs; //vector of proned HAs
    std::vector<Atomic> pronedFAs; //vector of proned FAs

    std::vector<Atomic>::iterator itHA = atomicsPair.first.begin();
    bool removeHA = false;
    while (itHA != atomicsPair.first.end())
    {
        removeHA = false;
        if (FAs_nodes_trace[(*itHA).outputs[0]] != 0 && FAs_nodes_trace[(*itHA).outputs[1]] != 0) //if HA's outputs are part of an FA
        {
            removeHA = true;
            if (!FAs_outputs_trace[(*itHA).outputs[0]]) //if HA's output[0] is not an FA's output
            {
                for (auto i : (*itHA).fanoutsVec[0])
                {
                    if (FAs_nodes_trace[(*itHA).outputs[0]] != FAs_nodes_trace[i])
                    {
                        removeHA = false;
                        FA_remove_indexes[FAs_nodes_indexes[i]] = true;
                        break;
                    }
                }
            }
            if (!FAs_outputs_trace[(*itHA).outputs[1]]) //if HA's output[0] is not an FA's output
            {
                for (auto i : (*itHA).fanoutsVec[1])
                {
                    if (FAs_nodes_trace[(*itHA).outputs[1]] != FAs_nodes_trace[i])
                    {
                        removeHA = false;
                        FA_remove_indexes[FAs_nodes_indexes[i]] = true;
                        break;
                    }
                }
            }
        }
        if (!removeHA)
            pronedHAs.push_back(*itHA);
        itHA++;
    }
    //swaping pronedHAs and the HAs vectors
    atomicsPair.first = pronedHAs;

    //removing FAs with more than two outputs (one of their HAs has an output to outside of the FA)
    for (int index=0; index<atomicsPair.second.size(); index++)
    {
        if (!FA_remove_indexes[index])
            pronedFAs.push_back(atomicsPair.second[index]);
    }
    
    //swaping pronedFAs and the FAs vectors
    atomicsPair.second = pronedFAs;
}

//***************************************************************************
void Atomic::removeAtomicsWithIntersection(std::vector<Atomic> &atomics)
{
    std::vector<Atomic> pronedAtomics;

    //to save all atomic blocks indexes that should be removed
    std::unordered_map<int, bool> removeList;

    //a hash table to map each node to the atomic blocks which contain it
    std::unordered_map<uint32_t, std::vector<int>> map_node_Atomic_position;

    for (int index=0; index<atomics.size(); index++)
    {
        for (auto i: atomics[index].nodeIndexList)
        {
            map_node_Atomic_position[i].push_back(index);
        }
    }

    //searching for nodes which are contained in more than one HA
    for (auto i: map_node_Atomic_position)
    {
        if (i.second.size() > 1)
        {
            for (auto j: i.second)
                removeList[j] = true;
        }
    }

    for (int index=0; index<atomics.size(); index++)
    {
        if (!removeList[index])
            pronedAtomics.push_back(atomics[index]);
    }

    atomics = pronedAtomics;

}