#include "cut.hpp"

Cutt::Cutt(std::vector<uint32_t> inputs, uint32_t output, std::vector<uint32_t> nodeIndexList, std::string truthTable, std::vector<uint32_t> fanouts)
{
    this->inputs = inputs;
    sort(this->inputs.begin(), this->inputs.end());
    this->nodeIndexList = nodeIndexList;
    sort(this->inputs.begin(), this->inputs.end());
    this->output = output;
    this->truthTable = truthTable;
    for (auto i : this->inputs)
        this->inputSig += std::to_string(i) + "-";
    this->fanouts = fanouts;
}

//***************************************************************************************
Cutt::Cutt(mockturtle::cut<16, mockturtle::cut_data<true, mockturtle::empty_cut_data>> cut, mockturtle::node<mockturtle::aig_network> rootNode, std::string truthTable, const mockturtle::aig_network &aig, mockturtle::fanout_view<mockturtle::aig_network> view_fanout)
{
    this->output = aig.node_to_index(rootNode); //the root node is also the output of the cut
    std::vector<mockturtle::node<mockturtle::aig_network>> leaves;
    for (auto j : cut) //iterating through cut to obain all input indexes
    {
        leaves.push_back(aig.index_to_node(j));
        this->inputs.push_back(j);
    }
    //creating a cut_view on the cut root node and the leaves
    //With the help of a cut_view, we can iterate through the cut nodes using foreach_node
    mockturtle::cut_view<mockturtle::aig_network> cutView{aig, leaves, rootNode};

    uint32_t nodeIndex;
    cutView.foreach_node([&](auto node) {
        nodeIndex = aig.node_to_index(node);
        //we should shouldn't add the 0 index
        //we should shouldn't add the input node indexes as they are not cut members
        if (nodeIndex != 0 && std::find(this->inputs.begin(), this->inputs.end(), nodeIndex) == this->inputs.end())
            this->nodeIndexList.push_back(nodeIndex); //extracting each node index and push it into nodeIndexList
    });

    //finding all cut fanouts
    view_fanout.foreach_fanout(rootNode, [&](const auto &p) {
        this->fanouts.push_back(aig.node_to_index(p));
    });

    //creating the input signature
    for (auto i : this->inputs)
        this->inputSig += std::to_string(i) + "-";
    this->truthTable = truthTable;
}

//***************************************************************************************
//we also consider truth tables after negating the inputs
CutType Cutt::identifyCut() const
{
    if (this->inputs.size() == 2)
    {
        if (truthTable == "9" || truthTable == "6")
            return HA_S;
        else if (truthTable == "1" || truthTable == "2" || truthTable == "4" || truthTable == "8" || truthTable == "e" || truthTable == "d" || truthTable == "b" || truthTable == "7")
            return HA_C;
    }
    else if (this->inputs.size() == 3)
    {
        if (truthTable == "96" || truthTable == "69")
            return FA_S;
        else if (truthTable == "e8" || truthTable == "8e" || truthTable == "17" || truthTable == "71" || truthTable == "4d" || truthTable == "d4" || truthTable == "2b" || truthTable == "b2")
            return FA_C;
    }
    return Unknown;
}
//***************************************************************************************

std::vector<uint32_t> Cutt::returnInputs() const
{
    return this->inputs;
}
//***************************************************************************************

std::uint32_t Cutt::returnOutput() const
{
    return this->output;
}
//***************************************************************************************

std::string Cutt::returnTruthTable() const
{
    return this->truthTable;
}
//***************************************************************************************

std::vector<uint32_t> Cutt::returnNodeIndexes() const
{
    return this->nodeIndexList;
}

//***************************************************************************************
std::string Cutt::returnInputSig() const
{
    return this->inputSig;
}

//***************************************************************************************
std::ostream &operator<<(std::ostream &output, const Cutt &C)
{
    output << C.output << " ---> ";
    for (auto i : C.inputs)
    {
        output << i << " ";
    }
    output << "[ ";
    for (auto i : C.nodeIndexList)
    {
        output << i << " ";
    }
    output << "]";
    output << " ===> " << C.truthTable;
    return output;
}

//***************************************************************************************
std::vector<Cutt *> Cutt::ExtractCuts(const mockturtle::aig_network &aig, uint32_t cut_size, uint32_t cut_limit)
{
    mockturtle::cut_enumeration_params cp;
    cp.cut_size = cut_size;   //maximum number of leaves
    cp.cut_limit = cut_limit; //maximum number of computed cuts for a node
    std::vector<Cutt *> aigCuts;
    std::string truthTable;

    //create a fanout_view on AIG so then we can iterate through each node's fanouts
    mockturtle::fanout_view<mockturtle::aig_network> view_fanout{aig};

    //cut enumeration algorithm from mockturtle
    auto cuts = mockturtle::cut_enumeration<mockturtle::aig_network, true>(aig, cp);

    //iterating through each node and extract their cuts
    aig.foreach_node([&](auto rootNode) {
        const auto index = aig.node_to_index(rootNode);
        for (auto const &cut : cuts.cuts(index))
        {
            truthTable = kitty::to_hex(cuts.truth_table(*cut));
            Cutt *c = new Cutt(*cut, rootNode, truthTable, aig, view_fanout);
            aigCuts.push_back(c);
        }
    });
    return aigCuts;
}